import tqdm

import urllib.request
import subprocess
import sys

from spotdl.encode.encoders import EncoderFFmpeg
from spotdl.metadata.embedders import EmbedderDefault
from spotdl.metadata import BadMediaFileError

import spotdl.util

import logging
logger = logging.getLogger(__name__)

CHUNK_SIZE = 16 * 1024

class Track:
    """
    This class allows for various operations on provided track
    metadata.

    Parameters
    ----------
    metadata: `dict`
        Track metadata in standardized form.

    cache_albumart: `bool`
        Whether or not to cache albumart data by making network request
        to the given URL. This caching is done as soon as a
        :class:`Track` object is created.

    Examples
    --------
    + Downloading the audio track *"NCS - Spectre"* in opus format from
      YouTube while simultaneously encoding it to an mp3 format:

        >>> from spotdl.metadata_search import MetadataSearch
        >>> provider = MetadataSearch("ncs spectre")
        >>> metadata = provider.on_youtube()
        # The same metadata can also be retrived using `ProviderYouTube`:
        >>> # from spotdl.metadata.providers import ProviderYouTube
        >>> # provider = ProviderYouTube()
        >>> # metadata = provider.from_query("ncs spectre")
        # However, it is recommended to use `MetadataSearch` whenever
        # possible as it provides a higher level API.
        >>>
        >>> from spotdl.track import Track
        >>> track = Track(metadata)
        >>> stream = metadata["streams"].get(
        ...     quality="best",
        ...     preftype="opus",
        ... )
        >>>
        >>> import spotdl.metadata
        >>> filename = spotdl.metadata.format_string(
        ...     "{artist} - {track-name}.{output-ext}",
        ...     metadata,
        ...     output_extension="mp3",
        ... )
        >>>
        >>> filename
        'NoCopyrightSounds - Alan Walker - Spectre [NCS Release].mp3'
        >>> track.download_while_re_encoding(stream, filename)
    """

    def __init__(self, metadata, cache_albumart=False):
        self.metadata = metadata
        self._chunksize = CHUNK_SIZE
        if cache_albumart:
            self._albumart_thread = self._cache_albumart()
        self._cache_albumart = cache_albumart

    def _cache_albumart(self):
        albumart_thread = spotdl.util.ThreadWithReturnValue(
            target=lambda url: urllib.request.urlopen(url).read(),
            args=(self.metadata["album"]["images"][0]["url"],)
        )
        albumart_thread.start()
        return albumart_thread

    def _calculate_total_chunks(self, filesize):
        """
        Determines the total number of chunks.

        Parameters
        ----------
        filesize: `int`
            Total size of file in bytes.

        Returns
        -------
        chunks: `int`
            Total number of chunks based on the file size and chunk
            size.
        """

        chunks = (filesize // self._chunksize) + 1
        return chunks

    def _make_progress_bar(self, iterations):
        """
        Creates a progress bar using :class:`tqdm`.

        Parameters
        ----------
        iterations: `int`
            Number of iterations to be performed.

        Returns
        -------
        progress_bar: :class:`tqdm.std.tqdm`
            An iterator object.
        """

        progress_bar = tqdm.trange(
            iterations,
            unit_scale=(self._chunksize // 1024),
            unit="KiB",
            dynamic_ncols=True,
            bar_format='{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt}KiB '
                '[{elapsed}<{remaining}, {rate_fmt}{postfix}]',
        )
        return progress_bar

    def download_while_re_encoding(self, stream, target_path, target_encoding=None,
                                   encoder=EncoderFFmpeg(must_exist=False), show_progress=True):
        """
        Downloads a stream while simuntaneously encoding it to a
        given target format.

        Parameters
        ----------
        stream: `dict`
            A `dict` containing stream information in the keys:
            `encoding`, `filesize`, `connection`.

        target_path: `str`
            Path to file to write the target stream to.

        target_encoding: `str`, `None`
            Specify a target encoding. If ``None``, the target encoding
            is automatically determined from the ``target_path``.

        encoder: :class:`spotdl.encode.EncoderBase` object
            A :class:`spotdl.encode.EncoderBase` object to use for encoding.

        show_progress: `bool`
            Whether or not to display a progress bar.
        """

        total_chunks = self._calculate_total_chunks(stream["filesize"])
        process = encoder.re_encode_from_stdin(
            stream["encoding"],
            target_path,
            target_encoding=target_encoding
        )
        response = stream["connection"]

        progress_bar = self._make_progress_bar(total_chunks)
        for _ in progress_bar:
            chunk = response.read(self._chunksize)
            process.stdin.write(chunk)

        process.stdin.close()
        process.wait()

    def download(self, stream, target_path, show_progress=True):
        """
        Downloads a stream.

        Parameters
        ----------
        stream: `dict`
            A `dict` containing stream information in the keys:
            `filesize`, `connection`.

        target_path: `str`
            Path to file to write the downloaded stream to.

        show_progress: `bool`
            Whether or not to display a progress bar.
        """

        total_chunks = self._calculate_total_chunks(stream["filesize"])
        progress_bar = self._make_progress_bar(total_chunks)
        response = stream["connection"]

        def writer(response, progress_bar, file_io):
            for _ in progress_bar:
                chunk = response.read(self._chunksize)
                file_io.write(chunk)

        write_to_stdout = target_path == "-"
        if write_to_stdout:
            file_io = sys.stdout.buffer
            writer(response, progress_bar, file_io)
        else:
            with open(target_path, "wb") as file_io:
                writer(response, progress_bar, file_io)

    def re_encode(self, input_path, target_path, target_encoding=None,
                  encoder=EncoderFFmpeg(must_exist=False), show_progress=True):
        """
        Encodes an already downloaded stream.

        Parameters
        ----------
        input_path: `str`
            Path to input file.

        target_path: `str`
            Path to target file.

        target_encoding: `str`
            Encoding to encode the input file to. If ``None``, the
            target encoding is determined from ``target_path``.

        encoder: :class:`spotdl.encode.EncoderBase` object
            A :class:`spotdl.encode.EncoderBase` object to use for encoding.

        show_progress: `bool`
            Whether or not to display a progress bar.
        """
        stream = self.metadata["streams"].getbest()
        total_chunks = self._calculate_total_chunks(stream["filesize"])
        process = encoder.re_encode_from_stdin(
            stream["encoding"],
            target_path,
            target_encoding=target_encoding
        )
        with open(input_path, "rb") as fin:
            for _ in tqdm.trange(total_chunks):
                chunk = fin.read(self._chunksize)
                process.stdin.write(chunk)

        process.stdin.close()
        process.wait()

    def apply_metadata(self, input_path, encoding=None, embedder=EmbedderDefault()):
        """
        Applies metadata on the audio file.

        Parameters
        ----------
        input_path: `str`
            Path to audio file to apply metadata to.

        encoding: `str`
            Encoding of the input audio file. If ``None``, the target
            encoding is determined from ``input_path``.

        embedder: :class:`spotdl.metadata.embedders.EmbedderDefault`
            An object of :class:`spotdl.metadata.embedders.EmbedderDefault`
            which depicts the metadata embedding strategy to use.
        """
        if self._cache_albumart:
            albumart = self._albumart_thread.join()
        else:
            albumart = None

        try:
            embedder.apply_metadata(
                input_path,
                self.metadata,
                cached_albumart=albumart,
                encoding=encoding,
            )
        except BadMediaFileError as e:
            msg = ("{} Such problems should be fixed "
                  "with FFmpeg set as the encoder.").format(e.args[0])
            logger.warning(msg)




# This code has been taken from PyTube
# https://github.com/nficano/pytube
# and is based on
# https://github.com/nficano/pytube/pull/643

# This is because the commit required to fix the issue
# https://github.com/nficano/pytube/issues/641
# hasn't been released on PyPI yet and is urgently
# required for spotdl to operate properly.

# The below code is LICENSED under:
"""
Copyright (c) 2019 Nick Ficano

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sub-license, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice, and every other copyright notice found in this
software, and all the attributions in every file, and this permission notice
shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
# The original LICENSE notice can be found on
# https://github.com/nficano/pytube/blob/master/LICENSE


import pytube

def apply_patches():
    pytube.__main__.apply_descrambler = apply_descrambler


# The below imports are required by the patch
import json
from urllib.parse import parse_qs, unquote

def apply_descrambler(stream_data, key):
    """Apply various in-place transforms to YouTube's media stream data.
    Creates a ``list`` of dictionaries by string splitting on commas, then
    taking each list item, parsing it as a query string, converting it to a
    ``dict`` and unquoting the value.
    :param dict stream_data:
        Dictionary containing query string encoded values.
    :param str key:
        Name of the key in dictionary.
    **Example**:
    >>> d = {'foo': 'bar=1&var=test,em=5&t=url%20encoded'}
    >>> apply_descrambler(d, 'foo')
    >>> print(d)
    {'foo': [{'bar': '1', 'var': 'test'}, {'em': '5', 't': 'url encoded'}]}
    """
    otf_type = "FORMAT_STREAM_TYPE_OTF"

    if key == "url_encoded_fmt_stream_map" and not stream_data.get(
        "url_encoded_fmt_stream_map"
    ):
        formats = json.loads(stream_data["player_response"])["streamingData"]["formats"]
        formats.extend(
            json.loads(stream_data["player_response"])["streamingData"][
                "adaptiveFormats"
            ]
        )
        try:
            stream_data[key] = [
                {
                    "url": format_item["url"],
                    "type": format_item["mimeType"],
                    "quality": format_item["quality"],
                    "itag": format_item["itag"],
                    "bitrate": format_item.get("bitrate"),
                    "is_otf": (format_item.get("type") == otf_type),
                }
                for format_item in formats
            ]
        except KeyError:
            cipher_url = []
            for data in formats:
                cipher = data.get("cipher") or data["signatureCipher"]
                cipher_url.append(parse_qs(cipher))
            stream_data[key] = [
                {
                    "url": cipher_url[i]["url"][0],
                    "s": cipher_url[i]["s"][0],
                    "type": format_item["mimeType"],
                    "quality": format_item["quality"],
                    "itag": format_item["itag"],
                    "bitrate": format_item.get("bitrate"),
                    "is_otf": (format_item.get("type") == otf_type),
                }
                for i, format_item in enumerate(formats)
            ]
    else:
        stream_data[key] = [
            {k: unquote(v) for k, v in parse_qsl(i)}
            for i in stream_data[key].split(",")
        ]




def format_string(string, metadata, output_extension="", sanitizer=lambda s: s):
    """
    Replaces any special tags contained in the string with their
    metadata values.

    Parameters
    ----------
    string: `str`
        A string containing any special tags.

    metadata: `dict`
        Metadata in standardized form.

    output_extension: `str`
        This is used to replace the special tag *"{output-ext}"* (if any)
        in the ``string`` passed.

    sanitizer: `function`
        This sanitizer function is called on every metadata value
        before replacing it with its special tag.

    Returns
    -------
    string: `str`
        A string with all special tags replaced with their
        corresponding metadata values.

    Examples
    --------
    + Formatting the string *"{artist} - {track-name}"* with metadata
      from YouTube:

        >>> from spotdl.metadata_search import MetadataSearch
        >>> searcher = MetadataSearch("ncs spectre")
        >>> metadata = searcher.on_youtube()
        >>> from spotdl.metadata import format_string
        >>> string = format_string("{artist} - {track-name}", metadata)
        >>> string
        'NoCopyrightSounds - Alan Walker - Spectre [NCS Release]'
    """

    formats = {
        "{track-name}"   : metadata["name"],
        "{artist}"       : metadata["artists"][0]["name"],
        "{album}"        : metadata["album"]["name"],
        "{album-artist}" : metadata["artists"][0]["name"],
        "{genre}"        : metadata["genre"],
        "{disc-number}"  : metadata["disc_number"],
        "{duration}"     : metadata["duration"],
        "{year}"         : metadata["year"],
        "{original-date}": metadata["release_date"],
        "{track-number}" : str(metadata["track_number"]).zfill(len(str(metadata["total_tracks"]))),
        "{total-tracks}" : metadata["total_tracks"],
        "{isrc}"         : metadata["external_ids"]["isrc"],
        "{track-id}"     : metadata.get("id", ""),
        "{output-ext}"   : output_extension,
    }

    for key, value in formats.items():
        string = string.replace(key, sanitizer(str(value)))

    return string




import subprocess
import os

from spotdl.encode import EncoderBase
from spotdl.encode.exceptions import EncoderNotFoundError
from spotdl.encode.exceptions import FFmpegNotFoundError

import logging
logger = logging.getLogger(__name__)


# Key: from format
# Subkey: to format
RULES = {
    "m4a": {
        "mp3": "-codec:v copy -codec:a libmp3lame -ar 48000",
        "opus": "-codec:a libopus -vbr on",
        "m4a": "-acodec copy",
        "flac": "-codec:a flac -ar 48000",
        "ogg": "-codec:a libvorbis -ar 48000",
        "opus": "-codec:a libopus -ar 48000",
    },
    "opus": {
        "mp3": "-codec:a libmp3lame -ar 48000",
        "m4a": "-cutoff 20000 -codec:a aac -ar 48000",
        "flac": "-codec:a flac -ar 48000",
        "ogg": "-codec:a libvorbis -ar 48000",
        "opus": "-acodec copy",
    },
}


class EncoderFFmpeg(EncoderBase):
    """
    A class for encoding media files using FFmpeg.

    Parameters
    ----------
    encoder_path: `str`
        Path to FFmpeg.

    must_exist: `bool`
        Error out immediately if the encoder isn't found in
        ``encoder_path``.

    Examples
    --------
    + Re-encode an OPUS stream from STDIN to an MP3:

        >>> import os
        >>> input_path = "audio.opus"
        >>> target_path = "audio.mp3"
        >>> input_path_size = os.path.getsize(input_path)
        >>>
        >>> from spotdl.encode.encoders import EncoderFFmpeg
        >>> ffmpeg = EncoderFFmpeg()
        >>> process = ffmpeg.re_encode_from_stdin(
        ...     input_encoding="opus",
        ...     target_path=target_path
        ... )
        >>>
        >>> chunk_size = 4096
        >>> total_chunks = (input_path_size // chunk_size) + 1
        >>>
        >>> with open(input_path, "rb") as fin:
        ...     for chunk_number in range(1, total_chunks+1):
        ...         chunk = fin.read(chunk_size)
        ...         process.stdin.write(chunk)
        ...         print("chunks encoded: {}/{}".format(
        ...             chunk_number,
        ...             total_chunks,
        ...         ))
        >>>
        >>> process.stdin.close()
        >>> process.wait()
    """

    def __init__(self, encoder_path="ffmpeg", must_exist=True):
        _loglevel = "-hide_banner -nostats -v warning"
        _additional_arguments = ["-b:a", "192k", "-vn"]
        try:
            super().__init__(encoder_path, must_exist, _loglevel, _additional_arguments)
        except EncoderNotFoundError as e:
            raise FFmpegNotFoundError(e.args[0])
        self._rules = RULES

    def set_trim_silence(self):
        self.set_argument("-af silenceremove=start_periods=1")

    def get_encoding(self, path):
        return super().get_encoding(path)

    def _generate_encoding_arguments(self, input_encoding, target_encoding):
        initial_arguments = self._rules.get(input_encoding)
        if initial_arguments is None:
            raise TypeError(
                'The input format ("{}") is not supported.'.format(
                input_encoding,
            ))
        arguments = initial_arguments.get(target_encoding)
        if arguments is None:
            raise TypeError(
                'The output format ("{}") is not supported.'.format(
                target_encoding,
            ))
        return arguments

    def set_debuglog(self):
        self._loglevel = "-loglevel debug"

    def _generate_encode_command(self, input_path, target_path,
                                 input_encoding=None, target_encoding=None):
        if input_encoding is None:
            input_encoding = self.get_encoding(input_path)
        if target_encoding is None:
            target_encoding = self.get_encoding(target_path)
        arguments = self._generate_encoding_arguments(
            input_encoding,
            target_encoding
        )
        command = [self.encoder_path] \
            + ["-y", "-nostdin"] \
            + self._loglevel.split() \
            + ["-i", input_path] \
            + arguments.split() \
            + self._additional_arguments \
            + ["-f", self.target_format_from_encoding(target_encoding)] \
            + [target_path]

        return command

    def re_encode(self, input_path, target_path, target_encoding=None, delete_original=False):
        encode_command = self._generate_encode_command(
            input_path,
            target_path,
            target_encoding=target_encoding
        )
        logger.debug("Calling FFmpeg with:\n{command}".format(
            command=encode_command,
        ))
        process = subprocess.Popen(encode_command)
        process.wait()
        encode_successful = process.returncode == 0
        if encode_successful and delete_original:
            os.remove(input_path)
        return process

    def re_encode_from_stdin(self, input_encoding, target_path, target_encoding=None):
        encode_command = self._generate_encode_command(
            "-",
            target_path,
            input_encoding=input_encoding,
            target_encoding=target_encoding,
        )
        logger.debug("Calling FFmpeg with:\n{command}".format(
            command=encode_command,
        ))
        process = subprocess.Popen(encode_command, stdin=subprocess.PIPE)
        return process




from abc import ABC
from abc import abstractmethod
from collections.abc import Sequence

class StreamsBase(Sequence):
    @abstractmethod
    def __init__(self, streams):
        """
        This method must parse audio streams into a list of
        dictionaries with the keys:
        "bitrate", "download_url", "encoding", "filesize".

        The list should typically be sorted in descending order
        based on the audio stream's bitrate.

        This sorted list must be assigned to ``self.streams``.
        """

        self.streams = streams

    def __repr__(self):
        return "Streams({})".format(self.streams)

    def __len__(self):
        return len(self.streams)

    def __getitem__(self, index):
        return self.streams[index]

    def __eq__(self, instance):
        return self.streams == instance.streams

    def getbest(self):
        """
        Returns the audio stream with the highest bitrate.
        """

        return self.streams[0]

    def getworst(self):
        """
        Returns the audio stream with the lowest bitrate.
        """

        return self.streams[-1]


class ProviderBase(ABC):
    def set_credentials(self, client_id, client_secret):
        """
        This method may or not be used depending on whether the
        metadata provider requires authentication or not.
        """

        raise NotImplementedError

    @abstractmethod
    def from_url(self, url):
        """
        Fetches metadata for the given URL.

        Parameters
        ----------
        url: `str`
            Media URL.

        Returns
        -------
        metadata: `dict`
            A *dict* of standardized metadata.
        """

        pass

    def from_query(self, query):
        """
        Fetches metadata for the given search query.

        Parameters
        ----------
        query: `str`
            Search query.

        Returns
        -------
        metadata: `dict`
            A *dict* of standardized metadata.
        """

        raise NotImplementedError

    @abstractmethod
    def _metadata_to_standard_form(self, metadata):
        """
        Transforms the metadata into a format consistent with all other
        metadata providers, for easy utilization.
        """

        pass




from mutagen.easyid3 import EasyID3
from mutagen.id3 import ID3, TORY, TYER, TPUB, APIC, USLT, COMM
from mutagen.mp4 import MP4, MP4Cover
from mutagen.flac import Picture, FLAC
from mutagen.oggvorbis import OggVorbis
from mutagen.oggopus import OggOpus

import urllib.request
import base64

from spotdl.metadata import EmbedderBase
from spotdl.metadata import BadMediaFileError

import logging
logger = logging.getLogger(__name__)

# Apple has specific tags - see mutagen docs -
# http://mutagen.readthedocs.io/en/latest/api/mp4.html
M4A_TAG_PRESET = {
    "album": "\xa9alb",
    "artist": "\xa9ART",
    "date": "\xa9day",
    "title": "\xa9nam",
    "year": "\xa9day",
    "originaldate": "purd",
    "comment": "\xa9cmt",
    "group": "\xa9grp",
    "writer": "\xa9wrt",
    "genre": "\xa9gen",
    "tracknumber": "trkn",
    "albumartist": "aART",
    "discnumber": "disk",
    "cpil": "cpil",
    "albumart": "covr",
    "copyright": "cprt",
    "tempo": "tmpo",
    "lyrics": "\xa9lyr",
    "comment": "\xa9cmt",
    "explicit": "rtng",
}

TAG_PRESET = {}
for key in M4A_TAG_PRESET.keys():
    TAG_PRESET[key] = key


class EmbedderDefault(EmbedderBase):
    """
    A class for applying metadata on media files.

    Examples
    --------
    - Applying metadata on an already downloaded MP3 file:

        >>> from spotdl.metadata_search import MetadataSearch
        >>> provider = MetadataSearch("ncs spectre")
        >>> metadata = provider.on_youtube()
        >>> from spotdl.metadata.embedders import EmbedderDefault
        >>> embedder = EmbedderDefault()
        >>> embedder.as_mp3("media.mp3", metadata)
    """
    supported_formats = ("mp3", "m4a", "flac", "ogg", "opus")

    def __init__(self):
        super().__init__()
        self._m4a_tag_preset = M4A_TAG_PRESET
        self._tag_preset = TAG_PRESET
        # self.provider = "spotify" if metadata["spotify_metadata"] else "youtube"
    def as_mp3(self, path, metadata, cached_albumart=None):
        """
        Apply metadata on MP3 media files.

        Parameters
        ----------
        path: `str`
            Path to the media file.

        metadata: `dict`
            Metadata (standardized) to apply to the media file.

        cached_albumart: `bool`
            An albumart image binary. If passed, the albumart URL
            present in the ``metadata`` won't be downloaded or used.
        """
        logger.debug('Writing MP3 metadata to "{path}".'.format(path=path))
        # EasyID3 is fun to use ;)
        # For supported easyid3 tags:
        # https://github.com/quodlibet/mutagen/blob/master/mutagen/easyid3.py
        # Check out somewhere at end of above linked file
        audiofile = EasyID3(path)
        self._embed_basic_metadata(audiofile, metadata, "mp3", preset=TAG_PRESET)
        audiofile["media"] = metadata["type"]
        audiofile["author"] = metadata["artists"][0]["name"]
        audiofile["lyricist"] = metadata["artists"][0]["name"]
        audiofile["arranger"] = metadata["artists"][0]["name"]
        audiofile["performer"] = metadata["artists"][0]["name"]
        provider = metadata["provider"]
        audiofile["website"] = metadata["external_urls"][provider]
        audiofile["length"] = str(metadata["duration"])
        if metadata["publisher"]:
            audiofile["encodedby"] = metadata["publisher"]
        if metadata["external_ids"]["isrc"]:
            audiofile["isrc"] = metadata["external_ids"]["isrc"]
        audiofile.save(v2_version=3)

        # For supported id3 tags:
        # https://github.com/quodlibet/mutagen/blob/master/mutagen/id3/_frames.py
        # Each class in the linked source file represents an id3 tag
        audiofile = ID3(path)
        if metadata["year"]:
            audiofile["TORY"] = TORY(encoding=3, text=metadata["year"])
            audiofile["TYER"] = TYER(encoding=3, text=metadata["year"])
        if metadata["publisher"]:
            audiofile["TPUB"] = TPUB(encoding=3, text=metadata["publisher"])
        provider = metadata["provider"]
        audiofile["COMM"] = COMM(
            encoding=3, text=metadata["external_urls"][provider]
        )
        if metadata["lyrics"]:
            audiofile["USLT"] = USLT(
                encoding=3, desc=u"Lyrics", text=metadata["lyrics"]
            )
        if cached_albumart is None:
            cached_albumart = urllib.request.urlopen(
                metadata["album"]["images"][0]["url"]
            ).read()
        try:
            audiofile["APIC"] = APIC(
                encoding=3,
                mime="image/jpeg",
                type=3,
                desc=u"Cover",
                data=cached_albumart,
            )
        except IndexError:
            pass

        audiofile.save(v2_version=3)

    def as_m4a(self, path, metadata, cached_albumart=None):
        """
        Apply metadata on FLAC media files.

        Parameters
        ----------
        path: `str`
            Path to the media file.

        metadata: `dict`
            Metadata (standardized) to apply to the media file.

        cached_albumart: `bool`
            An albumart image binary. If passed, the albumart URL
            present in the ``metadata`` won't be downloaded or used.
        """

        logger.debug('Writing M4A metadata to "{path}".'.format(path=path))
        # For supported m4a tags:
        # https://github.com/quodlibet/mutagen/blob/master/mutagen/mp4/__init__.py
        # Look for the class named `MP4Tags` in the linked source file
        audiofile = MP4(path)
        self._embed_basic_metadata(audiofile, metadata, "m4a", preset=M4A_TAG_PRESET)
        if metadata["year"]:
            audiofile[M4A_TAG_PRESET["year"]] = metadata["year"]
        provider = metadata["provider"]
        audiofile[M4A_TAG_PRESET["comment"]] = metadata["external_urls"][provider]
        if metadata["lyrics"]:
            audiofile[M4A_TAG_PRESET["lyrics"]] = metadata["lyrics"]
        # Explicit values: Dirty: 4, Clean: 2, None: 0
        audiofile[M4A_TAG_PRESET["explicit"]] = (4,) if metadata["explicit"] else (2,)
        try:
            if cached_albumart is None:
                cached_albumart = urllib.request.urlopen(
                    metadata["album"]["images"][0]["url"]
                ).read()
            audiofile[M4A_TAG_PRESET["albumart"]] = [
                MP4Cover(cached_albumart, imageformat=MP4Cover.FORMAT_JPEG)
            ]
        except IndexError:
            pass

        audiofile.save()

    def as_flac(self, path, metadata, cached_albumart=None):
        """
        Apply metadata on MP3 media files.

        Parameters
        ----------
        path: `str`
            Path to the media file.

        metadata: `dict`
            Metadata (standardized) to apply to the media file.

        cached_albumart: `bool`
            An albumart image binary. If passed, the albumart URL
            present in the ``metadata`` won't be downloaded or used.
        """

        logger.debug('Writing FLAC metadata to "{path}".'.format(path=path))
        # For supported flac tags:
        # https://github.com/quodlibet/mutagen/blob/master/mutagen/mp4/__init__.py
        # Look for the class named `MP4Tags` in the linked source file
        audiofile = FLAC(path)

        self._embed_basic_metadata(audiofile, metadata, "flac")
        self._embed_ogg_metadata(audiofile, metadata)
        self._embed_mbp_picture(audiofile, "metadata", cached_albumart, "flac")

        audiofile.save()

    def as_ogg(self, path, metadata, cached_albumart=None):
        logger.debug('Writing OGG Vorbis metadata to "{path}".'.format(path=path))
        audiofile = OggVorbis(path)

        self._embed_basic_metadata(audiofile, metadata, "ogg")
        self._embed_ogg_metadata(audiofile, metadata)
        self._embed_mbp_picture(audiofile, metadata, cached_albumart, "ogg")

        audiofile.save()

    def as_opus(self, path, metadata, cached_albumart=None):
        logger.debug('Writing Opus metadata to "{path}".'.format(path=path))
        audiofile = OggOpus(path)

        self._embed_basic_metadata(audiofile, metadata, "opus")
        self._embed_ogg_metadata(audiofile, metadata)
        self._embed_mbp_picture(audiofile, metadata, cached_albumart, "opus")

        audiofile.save()

    def _embed_ogg_metadata(self, audiofile, metadata):
        if metadata["year"]:
            audiofile["year"] = metadata["year"]
        provider = metadata["provider"]
        audiofile["comment"] = metadata["external_urls"][provider]
        if metadata["lyrics"]:
            audiofile["lyrics"] = metadata["lyrics"]

    def _embed_mbp_picture(self, audiofile, metadata, cached_albumart, encoding):
        image = Picture()
        image.type = 3
        image.desc = "Cover"
        image.mime = "image/jpeg"
        if cached_albumart is None:
            cached_albumart = urllib.request.urlopen(
                metadata["album"]["images"][0]["url"]
            ).read()
        image.data = cached_albumart

        if encoding == "flac":
            audiofile.add_picture(image)
        elif encoding == "ogg" or encoding == "opus":
            # From the Mutagen docs (https://mutagen.readthedocs.io/en/latest/user/vcomment.html)
            image_data = image.write()
            encoded_data = base64.b64encode(image_data)
            vcomment_value = encoded_data.decode("ascii")
            audiofile["metadata_block_picture"] = [vcomment_value]

    def _embed_basic_metadata(self, audiofile, metadata, encoding, preset=TAG_PRESET):
        audiofile[preset["artist"]] = metadata["artists"][0]["name"]
        if metadata["album"]["artists"][0]["name"]:
            audiofile[preset["albumartist"]] = metadata["album"]["artists"][0]["name"]
        if metadata["album"]["name"]:
            audiofile[preset["album"]] = metadata["album"]["name"]
        audiofile[preset["title"]] = metadata["name"]
        if metadata["release_date"]:
            audiofile[preset["date"]] = metadata["release_date"]
            audiofile[preset["originaldate"]] = metadata["release_date"]
        if metadata["genre"]:
            audiofile[preset["genre"]] = metadata["genre"]
        if metadata["copyright"]:
            audiofile[preset["copyright"]] = metadata["copyright"]
        if encoding == "flac" or encoding == "ogg" or encoding == "opus":
            audiofile[preset["discnumber"]] = str(metadata["disc_number"])
        else:
            audiofile[preset["discnumber"]] = [(metadata["disc_number"], 0)]
        zfilled_track_number = str(metadata["track_number"]).zfill(len(str(metadata["total_tracks"])))
        if encoding == "flac" or encoding == "ogg" or encoding == "opus":
            audiofile[preset["tracknumber"]] = zfilled_track_number
        else:
            if preset["tracknumber"] == TAG_PRESET["tracknumber"]:
                audiofile[preset["tracknumber"]] = "{}/{}".format(
                    zfilled_track_number, metadata["total_tracks"]
                )
            else:
                audiofile[preset["tracknumber"]] = [
                    (metadata["track_number"], metadata["total_tracks"])
                ]




import spotipy
import spotipy.oauth2 as oauth2

from spotdl.metadata import ProviderBase
from spotdl.metadata.exceptions import SpotifyMetadataNotFoundError

from spotdl.authorize.services import AuthorizeSpotify
from spotdl.authorize import SpotifyAuthorizationError
import spotdl.util

import logging
logger = logging.getLogger(__name__)


class ProviderSpotify(ProviderBase):
    """
    Fetch metadata using Spotify API in standardized form.

    Parameters
    ----------
    spotify: :class:`spotdl.authorize.services.AuthorizeSpotify`, :class:`spotipy.Spotify`, ``None``
        An authorized instance to make API calls to Spotify endpoints.

        If ``None``, it will attempt to reference an already created
        :class:`spotdl.authorize.services.AuthorizeSpotify` instance
        or you can set your own *Client ID* and *Client Secret*
        by calling :func:`ProviderSpotify.set_credentials` later on.

    Examples
    --------
    - Fetching a track's metadata using Spotify URI:

        >>> from spotdl.authorize.services import AuthorizeSpotify
        # It is necessary to authorize Spotify API otherwise API
        # calls won't pass through Spotify. That means we won't
        # be able to fetch metadata from Spotify.
        >>> AuthorizeSpotify(
        ...     client_id="your_spotify_client_id",
        ...     client_secret="your_spotify_client_secret",
        ... )
        >>>
        >>> from spotdl.metadata.providers import ProviderSpotify
        >>> provider = ProviderSpotify()
        >>> metadata = provider.from_url(
        ...     "https://open.spotify.com/track/0aTiUssEOy0Mt69bsavj6K"
        ... )
        >>> metadata["name"]
        'Descending'
    """

    def __init__(self, spotify=None):
        if spotify is None:
            try:
                spotify = AuthorizeSpotify()
            except SpotifyAuthorizationError:
                pass
        self.spotify = spotify

    def set_credentials(self, client_id, client_secret):
        """
        Set your own credentials to authorize with Spotify API.
        This is useful if you initially didn't authorize API calls
        while creating an instance of :class:`ProviderSpotify`.
        """
        token = self._generate_token(client_id, client_secret)
        self.spotify = spotipy.Spotify(auth=token)

    def assert_credentials(self):
        if self.spotify is None:
            raise SpotifyAuthorizationError(
                "You must first setup an AuthorizeSpotify instance, or pass "
                "in client_id and client_secret to the set_credentials method."
            )

    def from_url(self, url):
        self.assert_credentials()
        logger.debug('Fetching Spotify metadata for "{url}".'.format(url=url))
        metadata = self.spotify.track(url)
        return self._metadata_to_standard_form(metadata)

    def from_query(self, query):
        self.assert_credentials()
        tracks = self.search(query)["tracks"]["items"]
        if not tracks:
            raise SpotifyMetadataNotFoundError(
                'Spotify returned no tracks for the search query "{}".'.format(
                    query,
                )
            )
        return self._metadata_to_standard_form(tracks[0])

    def search(self, query):
        self.assert_credentials()
        return self.spotify.search(query)

    def _generate_token(self, client_id, client_secret):
        credentials = oauth2.SpotifyClientCredentials(
            client_secret=client_secret,
        )
        token = credentials.get_access_token()
        return token

    def _metadata_to_standard_form(self, metadata):
        self.assert_credentials()
        artist = self.spotify.artist(metadata["artists"][0]["id"])
        album = self.spotify.album(metadata["album"]["id"])

        try:
            metadata[u"genre"] = spotdl.util.titlecase(artist["genres"][0])
        except IndexError:
            metadata[u"genre"] = None
        try:
            metadata[u"copyright"] = album["copyrights"][0]["text"]
        except IndexError:
            metadata[u"copyright"] = None
        try:
            metadata[u"external_ids"][u"isrc"]
        except KeyError:
            metadata[u"external_ids"][u"isrc"] = None

        metadata[u"release_date"] = album["release_date"]
        metadata[u"publisher"] = album["label"]
        metadata[u"total_tracks"] = album["tracks"]["total"]

        # Some sugar
        metadata["year"], *_ = metadata["release_date"].split("-")
        metadata["duration"] = metadata["duration_ms"] / 1000.0
        metadata["provider"] = "spotify"

        # Remove unwanted parameters
        del metadata["duration_ms"]
        del metadata["available_markets"]
        del metadata["album"]["available_markets"]

        return metadata




import lyricwikia

from abc import ABC
from abc import abstractmethod


class LyricBase(ABC):
    """
    Defined lyric providers must inherit from this abstract base
    class and implement their own functionality for the below
    defined methods.
    """

    def from_url(self, url, linesep="\n", timeout=None):
        """
        Fetches lyrics given a URL.

        Parameters
        ----------
        url: `str`
            URL to fetch lyrics from.

        linesep: `str`
            Use this separator between every line of the lyrics.

        timeout: `int`, `None`
            Timeout duration such as if the server doesn't return a
            response in an expected time frame.

        Returns
        -------
        lyrics: `str`
        """
        raise NotImplementedError

    def from_artist_and_track(self, artist, track, linesep="\n", timeout=None):
        """
        Fetches lyrics given an artist and the track name.

        Parameters
        ----------
        artist: `str`
            Artist name.

        track: `str`
            Track name.

        linesep: `str`
            Use this separator between every line of the lyrics.

        timeout: `int`, `None`
            Timeout duration such as if the server doesn't return a
            response in an expected time frame.

        Returns
        -------
        lyrics: `str`
        """
        raise NotImplementedError

    def from_query(self, query, linesep="\n", timeout=None):
        """
        Fetches lyrics given a search query.

        Parameters
        ----------
        query: `str`
            A search query.

        linesep: `str`
            Use this separator between every line of the lyrics.

        timeout: `int`, `None`
            Timeout duration such as if the server doesn't return a
            response in an expected time frame.

        Returns
        -------
        lyrics: `str`
        """
        raise NotImplementedError




from spotdl.metadata.providers import ProviderSpotify
from spotdl.metadata.providers import ProviderYouTube
from spotdl.metadata.providers import YouTubeSearch
from spotdl.metadata.embedders import EmbedderDefault
from spotdl.metadata.exceptions import SpotifyMetadataNotFoundError
from spotdl.metadata.exceptions import BadMediaFileError
import spotdl.metadata

from spotdl.lyrics.providers import LyricWikia
from spotdl.lyrics.providers import Genius
from spotdl.lyrics.exceptions import LyricsNotFoundError

from spotdl.encode.encoders import EncoderFFmpeg

from spotdl.authorize.services import AuthorizeSpotify

from spotdl.track import Track
import spotdl.util
import spotdl.config

from spotdl.command_line.exceptions import NoYouTubeVideoFoundError
from spotdl.command_line.exceptions import NoYouTubeVideoMatchError
from spotdl.metadata_search import MetadataSearch

from spotdl.helpers.spotify import SpotifyHelpers
from spotdl.command_line.arguments import ArgumentHandler
import spotdl.helpers.exceptions

import sys
import os
import urllib.request

import logging
logger = logging.getLogger(__name__)

class Spotdl:
    """
    This class is directly involved with the command-line interface
    of the tool. It allows downloading of tracks, writing M3U
    playlists, and providers other useful methods.

    Parameters
    ----------
    args: `dict`
        A dictionary containing arguments. These passed arguments will
        override the default arguments used by the tool. In case an
        invalid combination of arguments is passed, an `ArgumentError`
        will be raised indicating the reason.

    Examples
    --------
    + To download a track:

        >>> from spotdl.command_line.core import Spotdl
        >>> args = {
        ...     "song": ["https://open.spotify.com/track/2lfPecqFbH8X4lHSpTxt8l",],
        ... }
        >>> with Spotdl(args) as spotdl_handler:
        ...     spotdl_handler.match_arguments()

    + To download tracks without metadata:

        >>> from spotdl.command_line.core import Spotdl
        >>> args = {
        ...     "song": [
        ...         "https://open.spotify.com/track/2lfPecqFbH8X4lHSpTxt8l",
        ...         "ncs spectre"
        ...     ],
        ...     "no_metadata": True,
        ... }
        >>> with Spotdl(args) as spotdl_handler:
        ...     spotdl_handler.match_arguments()

    Similary, you can pass additional arguments (refer to the config fiile
    to know what all arguments are supported).

    + To download tracks from file:

        >>> from spotdl.command_line.core import Spotdl
        >>> args = {
        ...     "list": "file_with_tracks.txt",
        ... }
        >>> with Spotdl(args) as spotdl_handler:
        ...     spotdl_handler.match_arguments()

    + You can also the call the download methods later on as per need:

        >>> from spotdl.command_line.core import Spotdl
        >>> args = {
        ...     "no_encode": True,
        ... }
        >>> with Spotdl(args) as spotdl_handler:
        ...     spotdl_handler.download_track("https://open.spotify.com/track/2lfPecqFbH8X4lHSpTxt8l")
        ...     print("Downloading 2nd track.")
        ...     spotdl_handler.download_track("ncs spectre")
        ...     print("Downloading from file.")
        ...     spotdl_handler.download_tracks_from_file("file_full_of_tracks.txt")

    Using `with` is optional. You can also directly create :class:`Spotdl` objects:

        >>> from spotdl.command_line.core import Spotdl
        >>> args = {
        ...     "no_encode": True,
        ... }
        >>> spotdl_handler = Spotdl(args)
        >>> spotdl_handler.download_track("ncs spectre")
    """

    def __init__(self, args={}):
        argument_handler = ArgumentHandler(args)
        arguments = argument_handler.run_errands()
        AuthorizeSpotify(
            client_id=arguments["spotify_client_id"],
            client_secret=arguments["spotify_client_secret"]
        )
        self.arguments = arguments

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        del self

    def match_arguments(self):
        """
        This is the main entry method that performs relevant work based
        on whatever arguments have been passed.
        """

        logger.debug("Received arguments:\n{}".format(self.arguments))

        if self.arguments.get("remove_config"):
            self.remove_saved_config()
            return 0
        self.save_default_config()

        spotify_tools = SpotifyHelpers()
        if self.arguments["song"]:
            for track in self.arguments["song"]:
                logger.info('Downloading "{}"'.format(track))
                if track == "-":
                    for line in sys.stdin:
                        self.download_track(
                            line.strip(),
                        )
                else:
                    self.download_track(track)
        elif self.arguments["list"]:
            if self.arguments["write_m3u"]:
                self.write_m3u(
                    self.arguments["list"],
                    self.arguments["write_to"]
                )
            else:
                list_download = {
                    "synchronous": self.download_tracks_from_file,
                    # "threaded"  : self.download_tracks_from_file_threaded,
                }[self.arguments["processor"]]

                list_download(
                    self.arguments["list"],
                )
        elif self.arguments["playlist"]:
            try:
                playlist = spotify_tools.fetch_playlist(self.arguments["playlist"])
            except spotdl.helpers.exceptions.PlaylistNotFoundError:
                return spotdl.command_line.exitcodes.URI_NOT_FOUND_ERROR
            else:
                spotify_tools.write_playlist_tracks(playlist, self.arguments["write_to"])
        elif self.arguments["album"]:
            try:
                album = spotify_tools.fetch_album(self.arguments["album"])
            except spotdl.helpers.exceptions.AlbumNotFoundError:
                return spotdl.command_line.exitcodes.URI_NOT_FOUND_ERROR
            else:
                spotify_tools.write_album_tracks(album, self.arguments["write_to"])
        elif self.arguments["all_albums"]:
            try:
                albums = spotify_tools.fetch_albums_from_artist(self.arguments["all_albums"])
            except spotdl.helpers.exceptions.ArtistNotFoundError:
                return spotdl.command_line.exitcodes.URI_NOT_FOUND_ERROR
            else:
                spotify_tools.write_all_albums(albums, self.arguments["write_to"])
        elif self.arguments["username"]:
            try:
                playlist_url = spotify_tools.prompt_for_user_playlist(self.arguments["username"])
            except spotdl.helpers.exceptions.UserNotFoundError:
                return spotdl.command_line.exitcodes.URI_NOT_FOUND_ERROR
            else:
                playlist = spotify_tools.fetch_playlist(playlist_url)
                spotify_tools.write_playlist_tracks(playlist, self.arguments["write_to"])

    def save_config(self, config_file=spotdl.config.DEFAULT_CONFIG_FILE, config=spotdl.config.DEFAULT_CONFIGURATION):
        """
        Writes provided configuration to config file.

        Parameters
        ----------
        config_file: `str`
            Path to write the configuration to.

        config: `dict`
            A `dict` consisting of configuration options.
        """

        config_dir = os.path.dirname(config_file)
        os.makedirs(config_dir, exist_ok=True)
        logger.info('Writing configuration to "{0}":'.format(config_file))
        spotdl.config.dump_config(config_file=config_file, config=spotdl.config.DEFAULT_CONFIGURATION)
        config = spotdl.config.dump_config(config=spotdl.config.DEFAULT_CONFIGURATION["spotify-downloader"])
        for line in config.split("\n"):
            if line.strip():
                logger.info(line.strip())
        logger.info(
            "Please note that command line arguments have higher priority "
            "than their equivalents in the configuration file.\n"
        )

    def save_default_config(self):
        """
        Writes the default configuration to the default config file if
        it does not already exist.
        """

        if not os.path.isfile(spotdl.config.DEFAULT_CONFIG_FILE):
            self.save_config()

    def remove_saved_config(self, config_file=spotdl.config.DEFAULT_CONFIG_FILE):
        """
        Removes the config file if it exists.

        Parameters
        ----------
        config_file: `str`
            Path to configuration file.
        """

        if os.path.isfile(spotdl.config.DEFAULT_CONFIG_FILE):
            logger.info('Removing "{}".'.format(spotdl.config.DEFAULT_CONFIG_FILE))
            os.remove(spotdl.config.DEFAULT_CONFIG_FILE)
        else:
            logger.info('File does not exist: "{}".'.format(spotdl.config.DEFAULT_CONFIG_FILE))

    def write_m3u(self, track_file, target_path=None):
        """
        Generates an M3U playlist from a given track file and writes it
        to a target file.

        Parameters
        ----------
        track_file: `str`
            Path to file consisting of tracks.

        target_path:`str`
            Path to file to write the M3U playlist to. `None` indicates
            to automatically determine it from the `track_file`.
        """

        with open(track_file, "r") as fin:
            tracks = fin.read().splitlines()

        logger.info(
            "Checking and removing any duplicate tracks in {}.".format(track_file)
        )
        # Remove duplicates and empty elements
        # Also strip whitespaces from elements (if any)
        tracks = spotdl.util.remove_duplicates(
            tracks,
            condition=lambda x: x,
            operation=str.strip
        )

        if target_path is None:
            target_path = "{}.m3u".format(track_file.split(".")[0])

        total_tracks = len(tracks)
        logger.info("Generating {0} from {1} YouTube URLs.".format(target_path, total_tracks))
        write_to_stdout = target_path == "-"
        m3u_headers = "#EXTM3U\n\n"
        if write_to_stdout:
            sys.stdout.write(m3u_headers)
        else:
            with open(target_path, "w") as output_file:
                output_file.write(m3u_headers)

        videos = []
        for n, track in enumerate(tracks, 1):
            search_metadata = MetadataSearch(
                track,
                lyrics=not self.arguments["no_metadata"],
                yt_search_format=self.arguments["search_format"],
                yt_manual=self.arguments["manual"]
            )
            try:
                video = search_metadata.best_on_youtube_search()
            except (NoYouTubeVideoFoundError, NoYouTubeVideoMatchError) as e:
                logger.error(e.args[0])
            else:
                logger.info(
                    "Matched track {0}/{1} ({2})".format(
                        str(n).zfill(len(str(total_tracks))),
                        total_tracks,
                        video["url"],
                    )
                )
                m3u_key = "#EXTINF:{duration},{title}\n{youtube_url}\n".format(
                    duration=spotdl.util.get_sec(video["duration"]),
                    title=video["title"],
                    youtube_url=video["url"],
                )
                logger.debug(m3u_key.strip())
                if write_to_stdout:
                    sys.stdout.write(m3u_key)
                else:
                    with open(target_path, "a") as output_file:
                        output_file.write(m3u_key)

    def download_track(self, track):
        """
        Downloads a track given a track query, Spotify URI or a YouTube
        URI.

        Parameters
        ----------
        track: `str`
            A Spotify URI, YouTube URI or a query.
        """

        subtracks = track.split("::")
        download_track = subtracks[0]
        custom_metadata_track = len(subtracks) > 1
        if custom_metadata_track:
            metadata_track = subtracks[1]
        else:
            metadata_track = download_track

        search_metadata = MetadataSearch(
            metadata_track,
            lyrics=not self.arguments["no_metadata"],
            yt_search_format=self.arguments["search_format"],
            yt_manual=self.arguments["manual"]
        )

        def threaded_metadata():
            try:
                if self.arguments["no_metadata"]:
                    metadata = search_metadata.on_youtube()
                else:
                    metadata = search_metadata.on_youtube_and_spotify()
            except (NoYouTubeVideoFoundError, NoYouTubeVideoMatchError) as e:
                logger.error(e.args[0])
            else:
                return metadata

        metadata = spotdl.util.ThreadWithReturnValue(target=threaded_metadata)
        metadata.start()
        if not custom_metadata_track:
            metadata = metadata.join()
            if not metadata:
                return
            return self.download_track_from_metadata(metadata)

        search_metadata = MetadataSearch(
            download_track,
            lyrics=False,
            yt_search_format=self.arguments["search_format"],
            yt_manual=self.arguments["manual"]
        )
        try:
            download_track_metadata = search_metadata.on_youtube()
        except (NoYouTubeVideoFoundError, NoYouTubeVideoMatchError) as e:
            logger.error(e.args[0])
            return

        metadata = metadata.join()
        if not metadata:
            return

        logger.info('Overriding metadata as per passed metadata-track'.format(metadata_track))
        metadata["streams"] = download_track_metadata["streams"]
        return self.download_track_from_metadata(metadata)

    def should_we_overwrite_existing_file(self, overwrite):
        """
        Returns a `boolean` based on the given value of `overwrite`
        parameter, where `overwrite` is one of `force`, `skip` or
        `prompt`. The method will prompt for input via STDIN if the
        value of `overwrite` is `prompt`.

        Parameters
        ----------
        overwrite: `str`
            One of `force`, `skip` or `prompt`.

        Returns
        -------
        to_overwrite: `bool`
            `True` or `False` depending on whether the existing file
            should be overwritten or not.
        """

        if overwrite == "force":
            logger.info("Forcing overwrite on existing file.")
            to_overwrite = True
        elif overwrite == "prompt":
            to_overwrite = input("Overwrite? (y/N): ").lower() == "y"
        else:
            logger.info("Not overwriting existing file.")
            to_overwrite = False

        return to_overwrite

    def download_track_from_metadata(self, metadata):
        """
        Downloads track audio from the provided metadata.

        Parameters
        ----------
        metadata: `dict`
            A `dict` consisting of metadata in standardized form.
        """

        track = Track(metadata, cache_albumart=(not self.arguments["no_metadata"]))
        stream = metadata["streams"].get(
            quality=self.arguments["quality"],
            preftype=self.arguments["input_ext"],
        )
        if stream is None:
            logger.error('No matching streams found for given input format: "{}".'.format(
                self.arguments["input_ext"]
            ))
            return

        if self.arguments["no_encode"]:
            output_extension = stream["encoding"]
        else:
            output_extension = self.arguments["output_ext"]

        filename = spotdl.metadata.format_string(
            self.arguments["output_file"],
            metadata,
            output_extension=output_extension,
            sanitizer=lambda s: spotdl.util.sanitize(
                s, spaces_to_underscores=self.arguments["no_spaces"]
            )
        )
        download_to_stdout = filename == "-"
        temp_filename = filename if download_to_stdout else filename + ".temp"

        to_skip_download = self.arguments["dry_run"]
        if os.path.isfile(filename):
            logger.info('A file with name "{filename}" already exists.'.format(
                filename=filename
            ))
            to_skip_download = to_skip_download \
                or not self.should_we_overwrite_existing_file(self.arguments["overwrite"])

        if to_skip_download:
            logger.debug("Skip track download.")
            return

        if not self.arguments["no_metadata"]:
            metadata["lyrics"].start()

        os.makedirs(os.path.dirname(filename) or ".", exist_ok=True)

        logger.info('Downloading to "{filename}"'.format(filename=filename))
        if self.arguments["no_encode"]:
            track.download(stream, temp_filename)
        else:
            encoder = EncoderFFmpeg()
            if self.arguments["trim_silence"]:
                encoder.set_trim_silence()
            track.download_while_re_encoding(
                stream,
                temp_filename,
                target_encoding=output_extension,
                encoder=encoder,
            )

        if not self.arguments["no_metadata"]:
            logger.info("Applying metadata")
            track.metadata["lyrics"] = track.metadata["lyrics"].join()
            self.apply_metadata(track, temp_filename, output_extension)

        if not download_to_stdout:
            logger.debug('Renaming "{temp_filename}" to "{filename}".'.format(
                temp_filename=temp_filename, filename=filename
            ))
            os.rename(temp_filename, filename)

        return filename

    def apply_metadata(self, track, filename, encoding=None):
        """
        Applies metadata to a given track. This is the same as calling
        :func:`spotdl.track.Track.apply_metadata` except this will
        only raise a warning if an unsupported output format has
        been passed unlike :func:`spotdl.track.Track.apply_metadata`
        which would raise a `TypeError`.

        Parameters
        ----------
        track: :class:`spotdl.track.Track` object
            A corresponding :class:`spotdl.track.Track` object.

        filename: `str`
            A filename where the audio file exists on the disk.

        encoding: `str`, `None`
            An encoding of `None` indicates to automatically determine
            it from the filename.
        """

        try:
            track.apply_metadata(filename, encoding=encoding)
        except MediaFileError:
            logger.warning("Cannot apply metadata on provided output format.")

    def strip_and_filter_duplicates(self, elements):
        """
        Removes any duplicate elements from the given list and strips
        any whitespaces from elements.

        Parameters
        ----------
        elements: `list`
            A list of elements.

        Returns
        -------
        filtered_elements: `list`
            A list of filtered elements.
        """

        filtered_elements = spotdl.util.remove_duplicates(
            elements,
            condition=lambda x: x,
            operation=str.strip
        )
        return filtered_elements

    def filter_against_skip_file(self, items, skip_file):
        """
        Removes the element from `items` if there already exists the
        same element in `skip_file`.

        Parameters
        ----------
        items: `list`
            A list of elements.

        skip_file: `str`
            Path to file.

        Returns
        -------
        filtered_items: `list`
            ``items`` but with duplicates removed.
        """

        skip_items = spotdl.util.readlines_from_nonbinary_file(skip_file)
        filtered_skip_items = self.strip_and_filter_duplicates(skip_items)
        filtered_items = [item for item in items if not item in filtered_skip_items]
        return filtered_items

    def download_tracks_from_file(self, path):
        """
        Download tracks from file.

        Parameters
        ----------
        path: `str`
            Path to the file consisting of tracks.
        """

        logger.info(
            'Checking and removing any duplicate tracks in "{}".'.format(path)
        )
        tracks = spotdl.util.readlines_from_nonbinary_file(path)
        tracks = self.strip_and_filter_duplicates(tracks)

        if self.arguments["skip_file"]:
            len_tracks_before = len(tracks)
            tracks = self.filter_against_skip_file(tracks, self.arguments["skip_file"])
            logger.info("Skipping {} tracks due to matches in skip file.".format(
                len_tracks_before - len(tracks))
            )
        # Overwrite file
        spotdl.util.writelines_to_nonbinary_file(path, tracks)

        logger.info(
            "Downloading {n} tracks.\n".format(n=len(tracks))
        )

        for position, track in enumerate(tracks, 1):
            log_track_query = '{position}. Downloading "{track}"'.format(
                position=position,
                track=track
            )
            logger.info(log_track_query)
            try:
                self.download_track(track)
            except (urllib.request.URLError, TypeError, IOError) as e:
                logger.exception(e.args[0])
                logger.warning(
                    "Failed to download current track due to possible network issue. "
                    "Will retry after other songs."
                )
                tracks.append(track)
            except KeyboardInterrupt:
                # The current track hasn't been downloaded completely.
                # Make sure we continue from here the next the program runs.
                tracks.insert(0, track)
                raise
            else:
                if self.arguments["write_successful_file"]:
                    with open(self.arguments["write_successful_file"], "a") as fout:
                        fout.write("{}\n".format(track))
            finally:
                spotdl.util.writelines_to_nonbinary_file(path, tracks[position:])
                print("", file=sys.stderr)

    """
    def download_tracks_from_file_threaded(self, path):
        # FIXME: Can we make this function cleaner?

        logger.info(
            "Checking and removing any duplicate tracks in {}.\n".format(path)
        )
        with open(path, "r") as fin:
            # Read tracks into a list and remove any duplicates
            tracks = fin.read().splitlines()

        # Remove duplicates and empty elements
        # Also strip whitespaces from elements (if any)
        spotdl.util.remove_duplicates(
            tracks,
            condition=lambda x: x,
            operation=str.strip
        )

        # Overwrite file
        with open(path, "w") as fout:
            fout.writelines(tracks)

        tracks_count = len(tracks)
        current_iteration = 1

        next_track = tracks.pop(0)
        metadata = {
            "current_track": None,
            "next_track": spotdl.util.ThreadWithReturnValue(
                target=search_metadata,
                args=(next_track, self.arguments["search_format"])
            )
        }
        metadata["next_track"].start()
        while tracks_count > 0:
            metadata["current_track"] = metadata["next_track"].join()
            metadata["next_track"] = None
            try:
                print(tracks_count, file=sys.stderr)
                print(tracks, file=sys.stderr)
                if tracks_count > 1:
                    current_track = next_track
                    next_track = tracks.pop(0)
                    metadata["next_track"] = spotdl.util.ThreadWithReturnValue(
                        target=search_metadata,
                        args=(next_track, self.arguments["search_format"])
                    )
                    metadata["next_track"].start()

                log_track_query = str(current_iteration) + ". {artist} - {track-name}"
                logger.info(log_track_query)
                if metadata["current_track"] is None:
                    logger.warning("Something went wrong. Will retry after downloading remaining tracks.")
                    pass
                print(metadata["current_track"]["name"], file=sys.stderr)
                # self.download_track_from_metadata(metadata["current_track"])
            except (urllib.request.URLError, TypeError, IOError) as e:
                print("", file=sys.stderr)
                logger.exception(e.args[0])
                logger.warning("Failed. Will retry after other songs\n")
                tracks.append(current_track)
            else:
                tracks_count -= 1
                if self.arguments["write_sucessful_file"]:
                    with open(self.arguments["write_sucessful_file"], "a") as fout:
                        fout.write(current_track)
            finally:
                current_iteration += 1
                with open(path, "w") as fout:
                    fout.writelines(tracks)
    """




import logging
import sys

from spotdl.command_line.core import Spotdl
from spotdl.command_line import arguments
from spotdl.command_line.exceptions import ArgumentError

import spotdl.command_line.exitcodes

# hardcode loglevel for dependencies so that they do not spew generic
# log messages along with spotdl.
logger = logging.getLogger(name=__name__)


def main():
    try:
        parser = arguments.get_arguments()
    except ArgumentError as e:
        logger.info(e.args[0])
        sys.exit(spotdl.command_line.exitcodes.ARGUMENT_ERROR)
    else:
        args = parser.parse_args().__dict__
    try:
        with Spotdl(args) as spotdl_handler:
            exitcode = spotdl_handler.match_arguments()
    except ArgumentError as e:
        parser.error(
            e.args[0],
            exitcode=spotdl.command_line.exitcodes.ARGUMENT_ERROR
        )
    except KeyboardInterrupt as e:
        print("", file=sys.stderr)
        logger.exception(e)
        sys.exit(spotdl.command_line.exitcodes.KEYBOARD_INTERRUPT)
    else:
        sys.exit(exitcode)


if __name__ == "__main__":
    main()




import pytube
from bs4 import BeautifulSoup
import json

import urllib.request
import threading
from collections.abc import Sequence

from spotdl.metadata import StreamsBase
from spotdl.metadata import ProviderBase
from spotdl.metadata.exceptions import YouTubeMetadataNotFoundError

import spotdl.util

import spotdl.patch
spotdl.patch.youtube.apply_patches()

import logging
logger = logging.getLogger(__name__)

BASE_SEARCH_URL = "https://www.youtube.com/results?sp=EgIQAQ%253D%253D&q={}"
HEADERS = [('Range', 'bytes=0-'),]


class YouTubeVideos(Sequence):
    """
    A class for dealing with YouTube videos.

    Parameters
    ----------
    videos: A `list` of `dict` video objects
    """

    def __init__(self, videos):
        self.videos = videos
        super().__init__()

    def __repr__(self):
        return "YouTubeVideos({})".format(self.videos)

    def __len__(self):
        return len(self.videos)

    def __getitem__(self, index):
        return self.videos[index]

    def __eq__(self, instance):
        return self.videos == instance.videos

    def bestmatch(self):
        """
        Returns
        -------
        video: `dict`
            A *dict* with the keys: *title*, *url* and *duration*.
        """

        video = self.videos[0]
        logger.debug("Matched with: {title} ({url}) [{duration}]".format(
            title=video["title"],
            url=video["url"],
            duration=video["duration"]
        ))
        return video


class YouTubeSearch:
    """
    Make query searches on YouTube and fetch resulting videos.

    Examples
    --------
    >>> from spotdl.metadata.providers.youtube import YouTubeSearch
    >>> youtube = YouTubeSearch()
    >>> results = youtube.search("python projects")
    >>> for result in results:
    ...     print(result["title"])
    ...     print(result["url"])
    ...     print(result["duration"])
    ...     print("")
    """

    def __init__(self):
        self.base_search_url = BASE_SEARCH_URL

    def generate_search_url(self, query):
        """
        Generates a search URL for YouTube for a given search query.

        Parameters
        ----------
        query: `str`
            The search query.

        Returns
        -------
        search_url: `str`
            A URL which corresponds to YouTube search results.
        """

        quoted_query = urllib.request.quote(query)
        search_url = self.base_search_url.format(quoted_query)
        return search_url

    def _fetch_response_html(self, url):
        response = urllib.request.urlopen(url)
        soup = BeautifulSoup(response.read(), "html.parser")
        return soup

    def _fetch_search_results_from_html(self, html, limit=10):
        videos = []
        for result in html:
            if not self._html_is_video(result):
                continue
            video = self._extract_video_details_from_html_result(result)
            videos.append(video)
            if len(videos) >= limit:
                break
        return videos

    def _extract_video_details_from_html_result(self, html):
        video_time = html.find("span", class_="video-time").get_text()
        inner_html = html.find("div", class_="yt-lockup-content")
        video_id = inner_html.find("a")["href"][-11:]
        video_title = inner_html.find("a")["title"]
        video_details = {
            "url": "https://www.youtube.com/watch?v=" + video_id,
            "title": video_title,
            "duration": video_time,
        }
        return video_details

    def _html_is_video(self, result):
        # ensure result is not a channel
        not_video = (
            result.find("channel") is not None
            or "yt-lockup-channel" in result.parent.attrs["class"]
            or "yt-lockup-channel" in result.attrs["class"]
        )
        # ensure result is not a mix/playlist
        not_video = not_video or "yt-lockup-playlist" in result.parent.attrs["class"]
        # ensure video result is not an advertisement
        not_video = not_video or result.find("googleads") is not None
        # ensure video result is not a live stream
        not_video = not_video or result.find("span", class_="video-time") is None
        video = not not_video
        return video

    def _fetch_search_results_from_json(self, json, limit=10):
        json = json["contents"]["twoColumnSearchResultsRenderer"]
        json = json["primaryContents"]["sectionListRenderer"]["contents"][0]
        json = json["itemSectionRenderer"]["contents"]
        videos = []
        for result in json:
            if not self._json_is_video(result):
                continue
            video = self._extract_video_details_from_json_result(result)
            videos.append(video)
            if len(videos) >= limit:
                break
        return videos

    def _extract_video_details_from_json_result(self, json):
        json = json["videoRenderer"]
        video_id = json["videoId"]
        video_title = json["title"]["runs"][0]["text"]
        video_time = json["lengthText"]["simpleText"]
        video_details = {
            "url": "https://www.youtube.com/watch?v=" + video_id,
            "title": video_title,
            "duration": video_time,
        }
        return video_details

    def _json_is_video(self, result):
        # ensure result is not a corrected search query
        video = "videoRenderer" in result
        return video

    def _fetch_search_results(self, html, limit=10):
        videos_html = html.find_all(
            "div", {"class": "yt-lockup-dismissable yt-uix-tile"}
        )
        if videos_html:
            return self._fetch_search_results_from_html(videos_html, limit=limit)

        # Sometimes YouTube can go crazy and return JSON data instead
        # of an HTML response. Handle such cases.
        logger.debug("YouTube returned malformed HTML. Attempting to parse possible JSON data.")

        html = str(html)
        search_start = 'window["ytInitialData"] = '
        videos_json_start = html.find(search_start) + len(search_start)
        search_end = "}}]}}}}}}"
        videos_json_end = videos_json_start + html[videos_json_start:].find(search_end) + len(search_end)
        videos_json = html[videos_json_start:videos_json_end] + "}"
        try:
            videos_json = json.loads(videos_json)
        except json.decoder.JSONDecodeError:
            return []

        return self._fetch_search_results_from_json(videos_json, limit=limit)

    def _is_server_side_invalid_response(self, videos, html):
        if videos:
            return False
        search_message = html.find("div", {"class":"search-message"})
        return search_message is None

    def search(self, query, limit=10, retries=2):
        """
        Search and scrape YouTube to return a list of matching videos.

        Parameters
        ----------
        query: `str`
            The search query.

        limit: `int`
            Return only first n results.

        retries: `int`
            YouTube can sometimes return invalid response. Maximum
            number of retries to make in such case.

        Returns
        -------
        videos: An instance of :class:`spotdl.metadata.providers.youtube.YouTubeVideos`.
        """

        search_url = self.generate_search_url(query)
        logger.debug('Fetching YouTube results for "{}" at "{}".'.format(query, search_url))
        html = self._fetch_response_html(search_url)
        videos = self._fetch_search_results(html, limit=limit)
        to_retry = retries > 0 and self._is_server_side_invalid_response(videos, html)
        if to_retry:
            logger.debug(
                "Retrying since YouTube returned invalid response for search "
                "results. Retries left: {retries}.".format(retries=retries)
            )
            return self.search(query, limit=limit, retries=retries-1)
        videos = YouTubeVideos(videos)
        return videos


class YouTubeStreams(StreamsBase):
    """
    A class for standardizing pytube YouTube streams.
    The standardized streams can be accessed by ``self.streams``,
    which is a list of streams.
    Each stream contains the keys: *bitrate*, *connection*,
    *download_url*, *encoding* and *filesize*.

    Parameters
    ----------
    streams: :class:`pytube.query.StreamQuery`
        PyTube streams.

    Examples
    --------
    >>> import pytube
    >>> video = pytube.YouTube("https://www.youtube.com/watch?v=XpmeVNxZ-Ks")
    >>> from spotdl.metadata.providers.youtube import YouTubeStreams
    >>> streams = YouTubeStreams(video.streams)
    >>> for stream in streams:
    ...     print(stream["encoding"], stream["filesize"])
    """

    def __init__(self, streams):
        self.network_headers = HEADERS
        audiostreams = streams.filter(only_audio=True).order_by("abr").desc()
        thread_pool = []
        self.streams = []
        for stream in audiostreams:
            encoding = "m4a" if "mp4a" in stream.audio_codec else stream.audio_codec
            standard_stream = {
                # Store only the integer part for bitrate. For example
                # the given bitrate would be "192kbps", we store only
                # the integer part (192) here and drop the rest.
                "bitrate": int(stream.abr[:-4]),
                "connection": None,
                "download_url": stream.url,
                "encoding": encoding,
                "filesize": None,
            }
            establish_connection = threading.Thread(
               target=self._store_connection,
               args=(standard_stream,),
            )
            thread_pool.append(establish_connection)
            establish_connection.start()
            self.streams.append(standard_stream)

        for thread in thread_pool:
            thread.join()

    def _store_connection(self, stream):
        response = self._make_request(stream["download_url"])
        stream["connection"] = response
        stream["filesize"] = int(response.headers["Content-Length"])

    def _make_request(self, url):
        request = urllib.request.Request(url)
        for header in self.network_headers:
            request.add_header(*header)
        return urllib.request.urlopen(request)

    def get(self, quality="best", preftype="automatic"):
        """
        Fetches a stream matching the given parameters.

        Parameters
        ----------
        quality: `str`
            Stream quality to fetch. Can either be "best" or "worst".

        preftype: `str`
            Stream encoding to prefer.

        Returns
        -------
        stream: `dict`, `None`
            The stream matching the given parameters. ``None`` if no
            matching stream was found.
        """

        if quality == "best":
            return self.getbest(preftype=preftype)
        elif quality == "worst":
            return self.getworst(preftype=preftype)
        else:
            return None

    def getbest(self, preftype="automatic"):
        """
        Fetches the best quality stream.

        Parameters
        ----------
        preftype: `str`
            Stream encoding to prefer.

        Returns
        -------
        stream: `dict`, `None`
            The stream matching the given parameters.
        """

        selected_stream = None
        if preftype == "automatic":
            selected_stream = self.streams[0]
        else:
            for stream in self.streams:
                if stream["encoding"] == preftype:
                    selected_stream = stream
                    break
        logger.debug('Selected best quality stream for "{preftype}" format:\n{stream}'.format(
            preftype=preftype,
            stream=selected_stream,
        ))
        return selected_stream

    def getworst(self, preftype="automatic"):
        """
        Fetches the worst quality stream.

        Parameters
        ----------
        preftype: `str`
            Stream encoding to prefer.

        Returns
        -------
        stream: `dict`, `None`
            The stream matching the given parameters.
        """

        selected_stream = None
        if preftype == "automatic":
            selected_stream = self.streams[-1]
        else:
            for stream in self.streams[::-1]:
                if stream["encoding"] == preftype:
                    selected_stream = stream
                    break
        logger.debug('Selected worst quality stream for "{preftype}" format:\n{stream}'.format(
            preftype=preftype,
            stream=selected_stream,
        ))
        return selected_stream


class ProviderYouTube(ProviderBase):
    """
    Fetch metadata from YouTube in standardized form.

    Examples
    --------
    + Making a YouTube search and fetching metadata for the first
      result

        >>> from spotdl.metadata.providers import ProviderYouTube
        >>> provider = ProviderYouTube()
        >>> metadata = provider.from_query("mitis - try (unreleased)")
        >>> metadata["name"]
        'MitiS - Try (Unreleased) ft RØRY'
    """

    def from_query(self, query, retries=5):
        watch_urls = self.search(query)
        if not watch_urls:
            raise YouTubeMetadataNotFoundError(
                'YouTube returned nothing for the given search '
                'query ("{}")'.format(query)
            )
        return self.from_url(watch_urls[0]["url"], retries=retries)

    def from_url(self, url, retries=5):
        logger.debug('Fetching YouTube metadata for "{url}".'.format(url=url))
        try:
            content = pytube.YouTube(url)
        except KeyError:
            # Sometimes YouTube can return unexpected response, in such a case
            # retry a few times before finally failing.
            if retries > 0:
                retries -= 1
                logger.debug(
                    "YouTube returned an unexpected response for "
                    "`pytube.YouTube({url})`. Retries left: {retries}".format(
                        url=url, retries=retries
                    )
                )
                return self.from_url(url, retries=retries)
            else:
                raise
        else:
            return self.from_pytube_object(content)

    def from_pytube_object(self, content):
        return self._metadata_to_standard_form(content)

    def search(self, query):
        return YouTubeSearch().search(query)

    def _fetch_publish_date(self, content):
        # XXX: This needs to be supported in PyTube itself
        # See https://github.com/nficano/pytube/issues/595
        position = content.watch_html.find("publishDate")
        publish_date = content.watch_html[position+16:position+25]
        return publish_date

    def _metadata_to_standard_form(self, content):
        publish_date = self._fetch_publish_date(content)
        metadata = {
            "name": content.title,
            "artists": [{"name": content.author}],
            "duration": content.length,
            "external_urls": {"youtube": content.watch_url},
            "album": {
                "images": [{"url": content.thumbnail_url}],
                "artists": [{"name": None}],
                "name": None,
            },
            "year": publish_date.split("-")[0],
            "release_date": publish_date,
            "type": "track",
            "disc_number": 1,
            "track_number": 1,
            "total_tracks": 1,
            "publisher": None,
            "external_ids": {"isrc": None},
            "lyrics": None,
            "copyright": None,
            "genre": None,
            "streams": YouTubeStreams(content.streams),
            "provider": "youtube",
        }
        return metadata



class PlaylistNotFoundError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class AlbumNotFoundError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class ArtistNotFoundError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class UserNotFoundError(Exception):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class SpotifyPlaylistNotFoundError(PlaylistNotFoundError):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class SpotifyAlbumNotFoundError(AlbumNotFoundError):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class SpotifyArtistNotFoundError(ArtistNotFoundError):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)


class SpotifyUserNotFoundError(UserNotFoundError):
    __module__ = Exception.__module__

    def __init__(self, message=None):
        super().__init__(message)




import shutil
import os

from abc import ABC
from abc import abstractmethod

from spotdl.encode.exceptions import EncoderNotFoundError

"""
  NOTE ON ENCODERS
  ================

* FFmeg encoders sorted in descending order based
  on the quality of audio produced:
  libopus > libvorbis >= libfdk_aac > aac > libmp3lame

* libfdk_aac encoder, due to copyrights needs to be compiled
  by end user on MacOS brew install ffmpeg --with-fdk-aac
  will do just that. Other OS? See:
  https://trac.ffmpeg.org/wiki/Encode/AAC

"""

_TARGET_FORMATS_FROM_ENCODING = {
    "m4a": "mp4",
    "mp3": "mp3",
    "opus": "opus",
    "flac": "flac",
    "ogg": "ogg",
    "opus": "opus"
}


class EncoderBase(ABC):
    """
    Defined encoders must inherit from this abstract base class
    and implement their own functionality for the below defined
    methods.
    """

    @abstractmethod
    def __init__(self, encoder_path, must_exist, loglevel, additional_arguments=[]):
        if must_exist and shutil.which(encoder_path) is None:
            raise EncoderNotFoundError(
                "{} executable does not exist or was not found in PATH.".format(
                    encoder_path
                )
            )
        self.encoder_path = encoder_path
        self._loglevel = loglevel
        self._additional_arguments = additional_arguments
        self._target_formats_from_encoding = _TARGET_FORMATS_FROM_ENCODING

    def set_argument(self, argument):
        """
        Set any arbitrary arguments which are passed when the encoder
        is invoked.

        Parameters
        ----------
        argument: `str`
        """
        self._additional_arguments += argument.split()

    def get_encoding(self, path):
        """
        Determine the encoding for a local audio file from its file
        extnension. Whether is it an "mp3", "wav", "m4a", etc.

        Parameters
        ----------
        path: `str`
            Path to media file.

        Returns
        -------
        encoding: `str`
        """
        _, extension = os.path.splitext(path)
        # Ignore the initial dot from file extension
        encoding = extension[1:]
        return encoding

    @abstractmethod
    def set_debuglog(self):
        """
        Enable verbose logging on the encoder.
        """
        pass

    @abstractmethod
    def _generate_encode_command(self, input_path, target_path):
        """
        This method must generate the complete command for that would
        be used to invoke the encoder and perform the encoding.
        """
        pass

    @abstractmethod
    def _generate_encoding_arguments(self, input_encoding, target_encoding):
        """
        This method must return the core arguments for the defined
        encoder such as defining the sample rate, audio bitrate,
        etc.
        """
        pass

    @abstractmethod
    def re_encode(self, input_path, target_path, target_encoding=None, delete_original=False):
        """
        Re-encode a given input file to a specified output file.

        Parameters
        ----------
        input_path: `str`
            Path to media file that needs re-encoding.

        target_path: `str`
            Path to output media file.

        target_encoding: `str`, `None`
            It maybe "mp3", "opus", etc. If ``None``, it is
            determined from the file extension passed in
            ``target_path``.

        delete_original: `bool`
            Whether or not to delete the original media file.
        """
        pass

    def target_format_from_encoding(self, encoding):
        """
        Determines the target stream format from given input encoding
        for use with encoder.

        Parameters
        ----------
        encoding: `str`

        Returns
        -------
        target_format: `str`
            Target format which can be accepted by most mainstream encoders.
        """
        target_format = self._target_formats_from_encoding[encoding]
        return target_format

    def re_encode_from_stdin(self, input_encoding, target_path, target_encoding=None):
        """
        Read a file from STDIN and re-encode it to a specified output
        file.

        Parameters
        ----------
        input_encoding: `str`
            Path to media file that needs re-encoding.

        target_path: `str`
            Path to output media file.

        target_encoding: `str`, `None`
            It maybe "mp3", "opus", etc. If ``None``, it is
            determined from the file extension passed in
            ``target_path``.
        """
        raise NotImplementedError




from spotdl.authorize.services import AuthorizeSpotify
import spotdl.helpers.exceptions
import spotdl.util

import sys
import spotipy

import logging
logger = logging.getLogger(__name__)

try:
    from slugify import SLUG_OK, slugify
except ImportError:
    logger.error("Oops! `unicode-slugify` was not found.")
    logger.info("Please remove any other slugify library and install `unicode-slugify`.")
    raise


class SpotifyHelpers:
    """
    Provides helper methods for accessing the Spotify API.

    Parameters
    ----------
    spotify: :class:`spotdl.authorize.services.AuthorizeSpotify`, :class:`spotipy.Spotify`, ``None``
        An authorized instance to make API calls to Spotify endpoints.

        If ``None``, it will attempt to reference an already created
        :class:`spotdl.authorize.services.AuthorizeSpotify` instance.
    """

    def __init__(self, spotify=None):
        if spotify is None:
            spotify = AuthorizeSpotify()
        self.spotify = spotify

    def prompt_for_user_playlist(self, username):
        """
        An interactive method that will display user's playlists
        and prompt to make a selection.

        Parameters
        ----------
        username: `str`
            Spotfiy username.

        Returns
        -------
        spotify_uri: `str`
            Spotify URI for the selected playlist.
        """
        playlists = self.fetch_user_playlist_urls(username)
        for i, playlist in enumerate(playlists, 1):
            playlist_details = "{0}. {1:<30}  ({2} tracks)".format(
                i, playlist["name"], playlist["tracks"]["total"]
            )
            print(playlist_details, file=sys.stderr)
        print("", file=sys.stderr)
        playlist = spotdl.util.prompt_user_for_selection(playlists)
        return playlist["external_urls"]["spotify"]

    def fetch_user_playlist_urls(self, username):
        """
        Fetches all user's playlists.

        Parameters
        ----------
        username: `str`
            Spotfiy username.

        Returns
        -------
        playlist_uris: `list`
            Containing all playlist URIs.
        """
        logger.debug('Fetching playlists for "{username}".'.format(username=username))
        try:
            playlists = self.spotify.user_playlists(username)
        except spotipy.client.SpotifyException:
            msg = ('Unable to find user "{}". Make sure the the user ID is correct '
                   'and then try again.'.format(username))
            logger.error(msg)
            raise spotdl.helpers.exceptions.SpotifyUserNotFoundError(msg)
        else:
            collected_playlists = []
            while True:
                for playlist in playlists["items"]:
                    # in rare cases, playlists may not be found, so playlists['next']
                    # is None. Skip these. Also see Issue #91.
                    if playlist["name"] is not None:
                        collected_playlists.append(playlist)
                if playlists["next"]:
                    playlists = self.spotify.next(playlists)
                else:
                    break
            return collected_playlists

    def fetch_playlist(self, playlist_uri):
        """
        Fetches playlist.

        Parameters
        ----------
        playlist_uri: `str`
            Spotify playlist URI.

        Returns
        -------
        playlist: `dict`
            Spotify API response object for the playlist endpoint.
        """
        logger.debug('Fetching playlist "{playlist}".'.format(playlist=playlist_uri))
        try:
            playlist = self.spotify.playlist(playlist_uri, fields="tracks,next,name")
        except spotipy.client.SpotifyException:
            msg = ('Unable to find playlist "{}". Make sure the the playlist ID is correct '
                   'and the playlist is set to publicly visible, and then try again.'.format(
                    playlist_uri
                  ))
            logger.error(msg)
            raise spotdl.helpers.exceptions.SpotifyPlaylistNotFoundError(msg)
        else:
            return playlist

    def write_playlist_tracks(self, playlist, target_path=None):
        """
        Writes playlist track URIs to file.

        Parameters
        ----------
        playlist: `dict`
            Spotify API response object for the playlist endpoint.

        target_path: `str`
            Write Spotify track URIs to this file.
        """
        tracks = playlist["tracks"]
        if not target_path:
            target_path = u"{0}.txt".format(slugify(playlist["name"], ok="-_()[]{}"))
        return self.write_tracks(tracks, target_path)

    def fetch_album(self, album_uri):
        """
        Fetches album.

        Parameters
        ----------
        album_uri: `str`
            Spotify album URI.

        Returns
        -------
        album: `dict`
            Spotify API response object for the album endpoint.
        """
        logger.debug('Fetching album "{album}".'.format(album=album_uri))
        try:
            album = self.spotify.album(album_uri)
        except spotipy.client.SpotifyException:
            msg = ('Unable to find album "{}". Make sure the album ID is correct '
                   'and then try again.'.format(album_uri))
            logger.error(msg)
            raise spotdl.helpers.exceptions.SpotifyAlbumNotFoundError(msg)
        else:
            return album

    def write_album_tracks(self, album, target_path=None):
        """
        Writes album track URIs to file.

        Parameters
        ----------
        album: `dict`
            Spotify API response object for the album endpoint.

        target_path: `str`
            Write Spotify track URIs to this file.
        """
        tracks = self.spotify.album_tracks(album["id"])
        if not target_path:
            target_path = u"{0}.txt".format(slugify(album["name"], ok="-_()[]{}"))
        return self.write_tracks(tracks, target_path)

    def fetch_albums_from_artist(self, artist_uri, album_type=None):
        """
        Fetches all Spotify albums from a Spotify artist in the US
        market.

        Parameters
        ----------
        artist_uri: `str`
            Spotify artist URI.

        album_type: `str`
            The type of album to fetch (ex: single) the default is
            all albums.

        Returns
        -------
        abums: `str`
            All albums received in the Spotify API response object.
        """
        logger.debug('Fetching all albums for "{artist}".'.format(artist=artist_uri))
        # fetching artist's albums limitting the results to the US to avoid duplicate
        # albums from multiple markets
        try:
            results = self.spotify.artist_albums(artist_uri, album_type=album_type, country="US")
        except spotipy.client.SpotifyException:
            msg = ('Unable to find artist "{}". Make sure the artist ID is correct '
                   'and then try again.'.format(artist_uri))
            logger.error(msg)
            raise spotdl.helpers.exceptions.SpotifyArtistNotFoundError(msg)
        else:
            albums = results["items"]
            # indexing all pages of results
            while results["next"]:
                results = self.spotify.next(results)
                albums.extend(results["items"])
            return albums

    def write_all_albums(self, albums, target_path=None):
        """
        Writes tracks from all albums into a file.

        Parameters
        ----------
        albums: `str`
            Spotfiy API response received in :func:`fetch_albums_from_artist`.

        target_path: `str`
            Write Spotify track URIs to this file.
        """
        # if no file if given, the default save file is in the current working
        # directory with the name of the artist
        if target_path is None:
            target_path = albums[0]["artists"][0]["name"] + ".txt"

        for album in albums:
            logger.info('Fetching album "{album}".'.format(album=album["name"]))
            self.write_album_tracks(album, target_path=target_path)

    def write_tracks(self, tracks, target_path):
        """
        Writes Spotify track URIs to file

        Parameters
        ----------
        tracks: `list`
            As returned in the Spotify API response.

        target_path: `str`
            Writes track URIs to this file.
        """
        def writer(tracks, file_io):
            track_urls = []
            while True:
                for item in tracks["items"]:
                    if "track" in item:
                        track = item["track"]
                    else:
                        track = item
                    # Spotify sometimes returns additional empty "tracks" in the API
                    # response. We need to discard such "tracks".
                    # See https://github.com/spotify/web-api/issues/1562
                    if track is None:
                        continue
                    try:
                        track_url = track["external_urls"]["spotify"]
                        file_io.write(track_url + "\n")
                        track_urls.append(track_url)
                    except KeyError:
                        # FIXME: Write "{artist} - {name}" instead of Spotify URI for
                        #        "local only" tracks.
                        logger.warning(
                            'Skipping track "{0}" by "{1}" (local only?)'.format(
                                track["name"], track["artists"][0]["name"]
                            )
                        )
                # 1 page = 50 results
                # check if there are more pages
                if tracks["next"]:
                    tracks = self.spotify.next(tracks)
                else:
                    break
            return track_urls

        logger.info(u"Writing {0} tracks to {1}.".format(tracks["total"], target_path))
        write_to_stdout = target_path == "-"
        if write_to_stdout:
            file_out = sys.stdout
            track_urls = writer(tracks, file_out)
        else:
            with open(target_path, "a") as file_out:
                track_urls = writer(tracks, file_out)
        return track_urls




import lyricwikia

from spotdl.lyrics.lyric_base import LyricBase
from spotdl.lyrics.exceptions import LyricsNotFoundError


class LyricWikia(LyricBase):
    """
    Fetch lyrics from LyricWikia.

    Examples
    --------
    + Fetching lyrics for *"Tobu - Cruel"*:

        >>> from spotdl.lyrics.providers import LyricWikia
        >>> genius = LyricWikia()
        >>> lyrics = genius.from_artist_and_track("Tobu", "Cruel")
        >>> print(lyrics)
    """

    def from_artist_and_track(self, artist, track, linesep="\n", timeout=None):
        """
        Returns the lyric string for the given artist and track.
        """
        try:
            lyrics = lyricwikia.get_lyrics(artist, track, linesep, timeout)
        except lyricwikia.LyricsNotFound as e:
            raise LyricsNotFoundError(e.args[0])
        return lyrics

